﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using MatrixLibrary;
using Tangra.SDK;

namespace Tangra.Addins
{
	public partial class frmMagnitudeFits : Form
	{
		private Dictionary<ITangraCatalogStar, int> m_Measurements = new Dictionary<ITangraCatalogStar, int>();
		private double m_Ka;
		private double m_Kb;
		private double m_Kc;

		private double m_Gamma, m_a, m_b;

		private Dictionary<ulong, List<int>> m_MeasurementsPerStar = new Dictionary<ulong, List<int>>();
		private List<ITangraCatalogStar> m_CatalogStars = new List<ITangraCatalogStar>();

		public frmMagnitudeFits(Dictionary<ulong, List<int>> measurementsPerStar, List<ITangraCatalogStar> catalogStars)
		{
			InitializeComponent();

			m_MeasurementsPerStar = measurementsPerStar;
			m_CatalogStars = catalogStars;
		}

		private void frmMagnitudeFits_Load(object sender, EventArgs e)
		{
            cbxColour.SelectedItem = 0;
            cbxCatalogMagnitude.SelectedItem = 0;

			LeastSquareFit();

			PlotGamma();
			PlotColour();
		}

		private string StandardGamma(double gamma)
		{
			if (gamma >= 0.32 && gamma < 0.37)
				return "(HI)";
			else if (gamma >= 0.42 && gamma < 0.47)
				return "(LO)";
			else if (gamma >= 0.98 && gamma < 1.02)
				return "(OFF)";
			else
				return string.Empty;
		}

        // Transformations to get B, V, Rc and Ic from J and K
        // Warner and Harris (2007)
        // http://brucegary.net/dummies/method0.html
        //
        // B =  J + 0.198  + 5.215  * (J-K) - 2.7785 * (J-K)2 + 1.7495 * (J-K)3 
        // V =  J + 0.1496 + 3.5143 * (J-K) - 2.325  * (J-K)2 + 1.4688 * (J-K)3
        // Rc = J + 0.1045 + 2.5105 * (J-K) - 1.7849 * (J-K)2 + 1.123  * (J-K)3
        // Ic = J + 0.0724 + 1.2816 * (J-K) - 0.4866 * (J-K)2 + 0.2963 * (J-K)3
        //
        // You can expect SE = 0.08, 0.05, 0.04 and 0.035 provided -0.1 < (J-K)< 1.0.
        //
        public static double GetBVfromJK(double jk)
        {
            return (0.198 - 0.1496) + (5.215 - 3.5143) * jk - (2.7785 - 2.325) * jk * jk + (1.7495 - 1.4688) * jk * jk * jk;
        }

		private void PlotGamma()
		{
            if (m_Measurements.Count > 2)
            {
                var xVals = new List<float>();
                var yVals = new List<float>();

                foreach (ITangraCatalogStar star in m_Measurements.Keys)
                {
                    int flux = m_Measurements[star];

                    double intensity = -2.5 * Math.Log10(flux);

                    double mInst = GetSelectedMagnitude(star) + (m_Kb / m_Ka) * GetSelectedColourIndex(star);

                    xVals.Add((float)intensity);
                    yVals.Add((float)mInst);
                }

                float xMin = xVals.Min();
                float xMax = xVals.Max();
                float xMargin = 3;
                float xCoeff = (picGamma.Width - 2 * xMargin) / (xMax - xMin);

                float yMargin = 3;
                float yMin = (float)(xMin * m_a + m_b);
                float yMax = (float)(xMax * m_a + m_b);
                float yCoeff = (picGamma.Height - 2 * yMargin) / (yMax - yMin);

                using (Graphics g = Graphics.FromImage(picGamma.Image))
                {
                    g.Clear(Color.DarkGray);

                    for (int i = 0; i < xVals.Count; i++)
                    {
                        float x = xMargin + (xVals[i] - xMin) * xCoeff;
                        float y = picGamma.Height - (yVals[i] - yMin) * yCoeff - yMargin;

                        g.FillEllipse(Brushes.Blue, x - 2, y - 2, 4, 4);
                    }

                    g.Save();
                }                
            }

            picGamma.Update();
		}

		private void PlotColour()
		{
            if (m_Measurements.Count > 2)
            {
                var xVals = new List<float>();
                var yVals = new List<float>();

                foreach (ITangraCatalogStar star in m_Measurements.Keys)
                {
                    int flux = m_Measurements[star];

                    double magDiff = -2.5 * Math.Log10(flux) * m_a + m_b - GetSelectedMagnitude(star);

                    double colorIdx = GetSelectedColourIndex(star);

                    xVals.Add((float)colorIdx);
                    yVals.Add((float)magDiff);
                }

                float xMin = xVals.Min();
                float xMax = xVals.Max();
                float xMargin = 3;
                float xCoeff = (picColourV.Width - 2 * xMargin) / (xMax - xMin);

                float yMargin = 3;
                float yMin = yVals.Min();
                float yMax = yVals.Max();
                float yCoeff = (picColourV.Height - 2 * yMargin) / (yMax - yMin);

                using (Graphics g = Graphics.FromImage(picColourV.Image))
                {
                    g.Clear(Color.DarkGray);

                    for (int i = 0; i < xVals.Count; i++)
                    {
                        float x = xMargin + (xVals[i] - xMin) * xCoeff;
                        float y = picColourV.Height - (yVals[i] - yMin) * yCoeff - yMargin;

                        g.FillEllipse(Brushes.Red, x - 2, y - 2, 4, 4);
                    }

                    g.Save();
                }
            }

            picColourV.Update();
		}

		private void btnPlot_Click(object sender, EventArgs e)
		{
			LeastSquareFit();

			PlotGamma();
			PlotColour();
		}

		private void LeastSquareFit()
		{
			m_Measurements = new Dictionary<ITangraCatalogStar, int>();

			if (m_MeasurementsPerStar.Count > 0)
			{
				foreach (ulong starNo in m_MeasurementsPerStar.Keys)
				{
					ITangraCatalogStar star = m_CatalogStars.FirstOrDefault(s => s.StarNo == starNo);

					float selectedMag = GetSelectedMagnitude(star);
					float selectedColourIndex = GetSelectedColourIndex(star);

					if (float.IsNaN(selectedMag) || float.IsNaN(selectedColourIndex))
						continue;

					List<int> measurements = m_MeasurementsPerStar[starNo];
					measurements.Sort();

					int median = -1;
					if (measurements.Count % 2 == 0)
						median = (measurements[measurements.Count / 2] + measurements[(measurements.Count / 2) - 1]) / 2;
					else
						median = measurements[measurements.Count / 2];

					m_Measurements.Add(star, median);
				}

				// Solve using least squares: -2.5 * a * Log(Median-Intensity) + b = A * Mv + B * Mjk + C;
                // -2.5 * a * Log(Median-Intensity) = A * Mv + B * Mjk + C - b

				if (m_Measurements.Count > 3)
				{
					Matrix Y = new Matrix(m_Measurements.Count, 1);
					Matrix X = new Matrix(m_Measurements.Count, 3);

					int i = 0;
					foreach (ITangraCatalogStar star in m_Measurements.Keys)
					{
						Y[i, 0] = -2.5 * Math.Log10(m_Measurements[star]);

						X[i, 0] = GetSelectedMagnitude(star);
						X[i, 1] = GetSelectedColourIndex(star);
						X[i, 2] = 1;

						i++;
					}

					Matrix x_T = Matrix.Transpose(X);
					Matrix xx = x_T * X;
					Matrix xx_inv = Matrix.Inverse(xx);
					Matrix av = (xx_inv * x_T) * Y;

					m_Ka = av[0, 0];
					m_Kb = av[1, 0];
					m_Kc = av[2, 0];


					// Solve using least squares: 2.5 * a * Log(Median-Intensity) + b = A * Mv + B * Mjk + C;

					// A / a = Ka; A = 1 (desired); a = 1 / Ka
					// B / a = Kb; B = Kb / Ka
					// (C - b) / a = Kc; C = 0 (desired); b = -Kc / Ka

					m_Gamma = m_Ka;
					m_a = 1 / m_Gamma;
                    m_b = -m_Kc / m_Ka;

                    lblGamma.Text = string.Format("Gamma = {0:0.00}{3}; MInstr = {1:0.00} * (2.5 * Log (Intensity)) {4} {2:0.00}", m_Gamma, m_a, Math.Abs(m_b), StandardGamma(m_Gamma), m_b < 0 ? "-" : "+");
                    lblVCoeff.Text = string.Format("MInstr - {2} = {1}{0:0.00} * ({3})", 
                            Math.Abs(m_Kb / m_Ka), m_Kb / m_Ka < 0 ? "-" : "+",
                            cbxCatalogMagnitude.SelectedItem, cbxColour.SelectedItem);

					picGamma.Image = new Bitmap(picGamma.Width, picGamma.Height);
					picColourV.Image = new Bitmap(picColourV.Width, picColourV.Height);				
					
				}
				else
				{
					m_Ka = float.NaN;
					m_Kb = float.NaN;
					m_Kc = float.NaN;
					m_Gamma = float.NaN;
					m_a = float.NaN;
					m_b = float.NaN;
				}
			}			
		}

		private float GetSelectedMagnitude(ITangraCatalogStar star)
		{
			if (cbxCatalogMagnitude.SelectedIndex == 0)
				return star.MagV;
			else if (cbxCatalogMagnitude.SelectedIndex == 1)
				return star.MagR;
			else
				return float.NaN;
		}

		private float GetSelectedColourIndex(ITangraCatalogStar star)
		{
            if (cbxColour.SelectedIndex == 0)
				return star.MagJ - star.MagK;
            else if (cbxColour.SelectedIndex == 1)
				return (float)GetBVfromJK(star.MagJ - star.MagK);
            else
				return float.NaN;
		}
	}
}
